import os
import posixpath

from django.core.exceptions import SuspiciousFileOperation
from django.utils.encoding import force_bytes


def to_bytes(content):
    """Wrap Django's force_bytes to pass through bytearrays."""
    if isinstance(content, bytearray):
        return content

    return force_bytes(content)


def clean_name(name):
    """
    Cleans the name so that Windows style paths work
    """
    # Normalize Windows style paths
    cleaned = posixpath.normpath(name).replace("\\", "/")

    # os.path.normpath() can strip trailing slashes so we implement
    # a workaround here.
    if name.endswith("/") and not cleaned.endswith("/"):
        # Add a trailing slash as it was stripped.
        cleaned = cleaned + "/"

    # Given an empty string, os.path.normpath() will return ., which we don't want
    if cleaned == ".":
        cleaned = ""

    return cleaned


def safe_join(base, *paths):
    """
    A version of django.utils._os.safe_join for S3 paths.

    Joins one or more path components to the base path component
    intelligently. Returns a normalized version of the final path.

    The final path must be located inside of the base path component
    (otherwise a ValueError is raised).

    Paths outside the base path indicate a possible security
    sensitive operation.
    """
    base_path = base
    base_path = base_path.rstrip("/")
    paths = [p for p in paths]

    final_path = base_path + "/"
    for path in paths:
        _final_path = posixpath.normpath(posixpath.join(final_path, path))
        # posixpath.normpath() strips the trailing /. Add it back.
        if path.endswith("/") or _final_path + "/" == final_path:
            _final_path += "/"
        final_path = _final_path
    if final_path == base_path:
        final_path += "/"

    # Ensure final_path starts with base_path and that the next character after
    # the base path is /.
    base_path_len = len(base_path)
    if not final_path.startswith(base_path) or final_path[base_path_len] != "/":
        raise ValueError("the joined path is located outside of the base path component")

    return final_path.lstrip("/")


def get_available_overwrite_name(name, max_length):
    """Truncate a filename to obey max_length limit"""
    if max_length is None or len(name) <= max_length:
        return name

    # Adapted from Django
    dir_name, file_name = os.path.split(name)
    file_root, file_ext = os.path.splitext(file_name)
    truncation = len(name) - max_length

    file_root = file_root[:-truncation]
    if not file_root:
        raise SuspiciousFileOperation(
            f'Storage tried to truncate away entire filename "{name}". Please make sure that the corresponding file field allows sufficient "max_length".'
        )
    return os.path.join(dir_name, f"{file_root}{file_ext}")



from datetime import timedelta

from django.conf import settings as django_settings
from django.core.exceptions import ImproperlyConfigured
from django.core.signals import setting_changed

DEFAULT_GZIP_CONTENT_TYPES = (
    "text/css",
    "text/javascript",
    "application/javascript",
    "application/x-javascript",
    "image/svg+xml",
)

DEFAULT_OBJECT_PARAMETERS = {}

DEFAULT_GCP_SETTINGS = {
    "project_id": None,
    "credentials": None,
}

DEFAULT_GCP_STORAGE_SETTINGS = {
    "bucket_name": None,
    "custom_endpoint": None,
    "location": "",
    "default_acl": None,
    "querystring_auth": True,
    "expiration": timedelta(seconds=86400),
    "gzip": False,
    "gzip_content_types": DEFAULT_GZIP_CONTENT_TYPES,
    "file_overwrite": True,
    "object_parameters": DEFAULT_OBJECT_PARAMETERS,
    "max_memory_size": 0,
    "blob_chunk_size": None,
}


class StorageSettings:
    """Combine GCP_ and GCP_STORAGE_ settings for a given store

    Settings are determined and cached. Initial determination is done lazily (on first setting access
    rather than on initialisation). This allows this class to be initialised prior to
    django's apps being ready.

    """

    def __init__(self, store_key, **overrides):
        setting_changed.connect(self._handle_settings_changed, dispatch_uid="ic" + str(id(self)))
        self._store_key = store_key
        self._cache = None
        self._overrides = overrides

    def __getattr__(self, setting_key):
        """Get a setting value"""

        if self._cache is None:
            self._update_settings()

        try:
            return self._cache[setting_key]
        except KeyError as e:
            raise AttributeError(f"No setting '{setting_key}' available in StorageSettings") from e

    @property
    def _stores_settings(self):
        """Get a complete dict of all stores defined in settings.py (media + static + extras)"""
        all_stores = {
            "media": getattr(django_settings, "GCP_STORAGE_MEDIA", None),
            "static": getattr(django_settings, "GCP_STORAGE_STATIC", None),
            **getattr(django_settings, "GCP_STORAGE_EXTRA_STORES", {}),
        }

        return dict((k, v) for k, v in all_stores.items() if v is not None)

    @property
    def _store_settings(self):
        """Dict of store settings defined in settings.py for the current store key"""
        try:
            return self._stores_settings[self._store_key]
        except KeyError as e:
            raise ImproperlyConfigured(
                f"Mismatch: specified store key '{self._store_key}' does not match 'media', 'static', or any store key defined in GCP_STORAGE_EXTRA_STORES"
            ) from e

    def _handle_settings_changed(self, **kwargs):
        self._update_settings()

    def _update_settings(self):
        """Fetch settings from django configuration, merge with defaults and cache

        Re-run on receiving ``setting_changed`` signal from django.
        """

        # Start with the default settings
        to_cache = {
            **DEFAULT_GCP_SETTINGS,
            **DEFAULT_GCP_STORAGE_SETTINGS,
        }

        # Add GCP_ settings from settings.py (common root level settings, used by the storage module)
        for setting_key in DEFAULT_GCP_SETTINGS.keys():  # pylint: disable=consider-iterating-dictionary
            try:
                to_cache[setting_key] = getattr(django_settings, f"GCP_{setting_key.upper()}")
            except AttributeError:
                # Not defined in django settings, don't add it to the dict
                pass

        # Add GCP_STORAGE_ settings (storage-specific settings for the current store)
        for setting_key in DEFAULT_GCP_STORAGE_SETTINGS.keys():  # pylint: disable=consider-iterating-dictionary
            try:
                to_cache[setting_key] = self._store_settings[setting_key]
            except KeyError:
                pass

        # Add overrides and place in cache
        self._cache = {
            **to_cache,
            **self._overrides,
        }

        self.check()

    def check(self):
        """Check the settings on this object"""
        if self.location.startswith("/"):
            correct = self.location.lstrip("/\\")
            raise ImproperlyConfigured(
                f"'location' option in GCP_STORAGE_ cannot begin with a leading slash. Found '{self.location}'. Use '{correct}' instead."
            )



from datetime import timedelta
from os.path import split


def get_path(instance, field_name):
    """Get the path of the blob in the object store"""
    field_value = getattr(instance, field_name)
    return field_value.get("path", None) if field_value is not None else None


def get_blob(instance, field_name, reload=True):
    """Get a blob from a model instance containing a BlobField

    This allows you to download the blob to a local file. For example:

    ```py
    blob = get_blob(mymodel, "my_field_name")

    logger.info("Downloading file %s from bucket %s", blob.name, blob.bucket.name)

    # Download the blob to the temporary directory
    blob_file_name = os.path.split(blob.name)[-1]
    blob.download_to_filename(blob_file_name)
    ```

    :param django.db.Model instance: An instance of a django Model which has a BlobField
    :param str field_name: The name of the BlobField attribute on the instance
    :param bool reload: Default True. If it is not essential to have up-to-date information from the store, speed up the call to get_blob using call with reload_blob=False
    """
    path = get_path(instance, field_name)
    if path is not None:
        field = instance._meta.get_field(field_name)
        blob = field.storage.bucket.blob(path)
        if reload:
            blob.reload()

        return blob


def get_blob_name(instance, field_name):
    """Get the name of the blob including its extension (absent any path)

    The name is the object path absent any folder prefixes,
    eg if blob is located at path `mystuff/1234/myfile.txt` the name
    is `myfile.txt`
    """
    path = get_path(instance, field_name)
    if path is not None:
        return split(path)[-1]


def get_signed_url(instance, field_name, expiration=None, **kwargs):
    """Get a signed URL to the blob for the given model field name
    :param str field_name: Name of the model field (which should be a BlobField)
    :param Union[datetime.datetime|datetime.timedelta|None] expiration: Expiration date or duration for the URL. If None, duration defaults to 24hrs.
    :return str: Signed URL of the blob
    """
    expiration = expiration or timedelta(hours=24)
    blob = get_blob(instance, field_name)
    if blob is not None:
        return blob.generate_signed_url(expiration=expiration, **kwargs)


def get_signed_download_url(instance, field_name, **kwargs):
    """Gets a signed URL with the response disposition set to an attachment"""
    name = get_blob_name(instance, field_name)
    if name is not None:
        return get_signed_url(instance, field_name, **kwargs, response_disposition=f"attachment; filename={name}")


def get_console_url(instance, field_name):
    """Gets the URL of the blob in the GCS console"""
    path = get_path(instance, field_name)
    if path is not None:
        field = instance._meta.get_field(field_name)
        bucket_name = field.storage.bucket.name
        return f"https://console.cloud.google.com/storage/browser/{bucket_name}/{path}"


class BlobFieldMixin:
    """Mixin to a model to provide extra utility methods for processing of blobs"""

    def get_blob(self, field_name):
        """Get a blob object for the given model field name"""
        return get_blob(self, field_name)

    def get_blob_name(self, field_name):
        """Get blob name for the given model field name"""
        return get_blob_name(self, field_name)

    @classmethod
    def get_bucket(cls, field_name):
        return cls._meta.get_field(field_name).storage.bucket

    @classmethod
    def get_bucket_name(cls, field_name):
        return cls.get_bucket(field_name).name

    def get_console_url(self, field_name):
        """Get a URL to where the file resides in GCP cloud console"""
        return get_console_url(self, field_name)

    def get_path(self, field_name):
        """Get the path of the blob in the object store for the given model field name"""
        return get_path(self, field_name)

    def get_signed_url(self, field_name, expiration=None):
        """Get a signed URL to the blob for the given model field name"""
        return get_signed_url(self, field_name, expiration)

    def get_signed_download_url(self, field_name, expiration=None):
        """Get a signed URL to the blob with the response disposition set"""
        return get_signed_download_url(self, field_name, expiration)

    # def override_path(self, field_name, path):
    #     """Set a blobfield path manually, avoiding any quality control, checks or ingress"""



import json
import logging
import os
from uuid import uuid4

from django.conf import settings
from django.contrib.admin.widgets import AdminTextareaWidget
from django.core import checks
from django.core.exceptions import ValidationError
from django.db import models, transaction
from django.utils.translation import gettext_lazy as _

from .forms import CloudObjectFormField
from .gcloud import GoogleCloudStorage
from .operations import UNLIMITED_MAX_SIZE, blob_exists, copy_blob, get_signed_upload_url
from .widgets import DEFAULT_ACCEPT_MIMETYPE, CloudObjectWidget

logger = logging.getLogger(__name__)

DEFAULT_OVERWRITE_MODE = "never"

OVERWRITE_MODES = [
    "never",
    "update",
    "update-versioned",
    "add",
    "add-versioned",
    "add-update",
    "add-update-versioned",
]

DEFAULT_OVERRIDE_BLOBFIELD_VALUE = False


class BlobField(models.JSONField):
    """A FileField replacement with cloud store object-specific features

    Features such as metadata fetching/synchronisation and direct uploads are enabled.

    BlobField allows storage classes to be declared in settings. This enables:
    - storages to be switched without database migrations (supporting integration
      and release preview type workflows plus cloud migration)
    - eaiser way of using different buckets for different BlobFields without
      defining a Storage class for each bucket

    BlobField is built on JSONField to allow more complex information to be stored
    and queried. Examples might include:
     - cached object generation or metageneration,
     - cached object metadata, or
     - status of direct uploads

    BlobField does not inherit directly from FileField to avoid the burden of strict
    compatibility with legacy or irrelevant-to-cloud-storage behaviour (like django
    admin overriding formfields or coping with edge cases around name and path handling).
    We do this to support a clear and maintainable implementation.

    :param ingress_to: A string defining the path within the bucket to which direct uploads
    will be ingressed. These uploaded files will be moved to their ultimate path in the store on save of the model.

    :param accept_mimetype: A string passed to widgets, suitable for use in the `accept` attribute
    of an html filepicker. This will allow you to narrow down, eg to `image/*` or an even more
    specific mimetype. No validation is done at the field level that objects actually are of this
    type (because the python mimetypes module doesn't accept wildcard mimetypes and we don't want to
    go on a wild goose chase just for this small feature).

    :param get_destination_path: A callable to determine the destination path of the blob.
    The blob should already have been successfully uploaded to the temporary location
    in GCP prior to creation of this model instance. This function is then called in the
    pre_save stage of the new model instance, so it has access to all the model fields in
    order to construct the destination filename. Note this EXCLUDES the id, if the model has an
    autoincrementing ID field that gets created on save, because that can't be known prior to
    the save.

    :param max_size_bytes: The maximum size in bytes of files that can be uploaded.

    :param overwrite_mode: One of `OVERWRITE_MODES` determining the circumstances under which overwrite
    is allowed. overwrite mode behaves as follows:
    - never: Never allows overwrite
    - update: Only when updating an object
    - update-versioned: Only when updating an object and the bucket has object versioning
    - add: Only when adding an object (we're not sure why you'd want this, here for completeness)
    - add-versioned: Only when adding an object to a versioned bucket (again, for completeness)
    - add-update: Always allow (ie when adding or updating an object)
    - add-update-versioned: When adding or updating an object to a versioned bucket

    :param on_change: A callable that will be executed on change of the field value. This will be called
    on commit of the transaction (ie once any file upload is ingressed to its final location) and allows you,
    for example, to dispatch a worker task to further process the uploaded blob.
    """

    description = _("A GCP Cloud Storage object")
    empty_values = [None, "", [], ()]

    def __init__(
        self,
        get_destination_path=None,
        ingress_to="_tmp/",
        store_key="media",
        accept_mimetype=DEFAULT_ACCEPT_MIMETYPE,
        overwrite_mode=DEFAULT_OVERWRITE_MODE,
        max_size_bytes=None,
        on_change=None,
        update_attributes=None,
        **kwargs,
    ):
        self._validated = False
        self._cleaned = False
        self._on_commit_blank = None
        self._on_commit_valid = None
        self._versioning_enabled = None
        self._temporary_path = None
        self._max_size_bytes = max_size_bytes
        self._primary_key_set_explicitly = "primary_key" in kwargs
        self._choices_set_explicitly = "choices" in kwargs
        self.overwrite_mode = overwrite_mode
        self.get_destination_path = get_destination_path
        self.update_attributes = update_attributes
        self.ingress_to = ingress_to
        self.ingress_path = None
        self.store_key = store_key
        self.accept_mimetype = accept_mimetype
        self.on_change = on_change
        kwargs["default"] = kwargs.pop("default", None)
        kwargs["help_text"] = kwargs.pop("help_text", "GCP cloud storage object")

        # Note, if you want to define overrides, then use the GCP_STORAGE_EXTRA_STORES
        # setting with a different key
        self.storage = GoogleCloudStorage(store_key=store_key)

        # We should consider if there's a good use case for customising the storage class:
        #
        # self.storage = import_string(storage_class)(store_key=store_key)
        #
        # If doing so we should also check that the resulting storage class is a subclass
        # of GoogleCloudStorage, to ensure that the storage API is compatible:
        #
        # if not isinstance(self.storage, GoogleCloudStorage):
        #     raise TypeError(
        #         "%s.storage must be a subclass/instance of %s.%s"
        #         % (
        #             self.__class__.__qualname__,
        #             GoogleCloudStorage.__module__,
        #             GoogleCloudStorage.__qualname__,
        #         )
        #     )
        super().__init__(**kwargs)

    def check(self, **kwargs):
        return [
            *super().check(**kwargs),
            *self._check_explicit(),
            *self._check_get_destination_path(),
            *self._check_update_attributes(),
            *self._check_ingress_to(),
            *self._check_overwrite_mode(),
            *self._check_on_change(),
        ]

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        kwargs["store_key"] = self.store_key
        kwargs["ingress_to"] = self.ingress_to
        kwargs["accept_mimetype"] = self.accept_mimetype
        kwargs["overwrite_mode"] = self.overwrite_mode
        kwargs["get_destination_path"] = self.get_destination_path
        kwargs["update_attributes"] = self.update_attributes
        kwargs["on_change"] = self.on_change
        return name, path, args, kwargs

    def formfield(self, **kwargs):
        widget = kwargs.pop("widget", None)
        if widget is None or issubclass(widget, AdminTextareaWidget):
            widget = CloudObjectWidget(
                accept_mimetype=self.accept_mimetype,
                signed_ingress_url=self._get_signed_ingress_url(),
                max_size_bytes=self.max_size_bytes,
                ingress_path=self._get_temporary_path(),
            )

        # Set up some defaults while letting the caller override them
        defaults = {"form_class": CloudObjectFormField, "widget": widget}
        defaults.update(kwargs)
        return super().formfield(**defaults)

    @property
    def override_blobfield_value(self):
        """Shortcut to access the GCP_STORAGE_OVERRIDE_BLOBFIELD_VALUE setting"""
        return getattr(
            settings,
            "GCP_STORAGE_OVERRIDE_BLOBFIELD_VALUE",
            DEFAULT_OVERRIDE_BLOBFIELD_VALUE,
        )

    @property
    def max_size_bytes(self):
        """Shortcut to determine the max size in bytes allowable as an upload

        Defaults to unlimited upload size (which is dangerous, but since you're
        signing a URL for the user to upload we assume you've established trust
        in the user and know what you're doing!).
        """
        if self._max_size_bytes is not None:
            return self._max_size_bytes

        return getattr(
            settings,
            "GCP_STORAGE_BLOBFIELD_MAX_SIZE_BYTES",
            UNLIMITED_MAX_SIZE,
        )

    def clean(self, value, model_instance, skip_validation=False):
        """Clean the value to ensure correct state and on_commit hooks.
        This method is called during clean_fields on model save.
        """

        # An escape hatch allowing you to set the path directly. This is dangerous and should only be done
        # explicitly for the purpose of data migration and manipulation. You should never allow an untrusted
        # client to set paths directly, because knowing the path of a pre-existing object allows you to assume
        # access to it. Tip: You can use django's override_settings context manager to set this temporarily.
        if getattr(
            settings,
            "GCP_STORAGE_OVERRIDE_BLOBFIELD_VALUE",
            DEFAULT_OVERRIDE_BLOBFIELD_VALUE,
        ):
            logger.warning(
                "Overriding %s value to %s",
                self.__class__.__name__,
                value,
            )
            new_value = value

        else:
            # Validate the value given
            if not skip_validation and not self._validated:
                self.validate(value, model_instance)

            # Clean {} --> None
            value = self._clean_blank_value(getattr(model_instance, self.attname))

            # Get model state
            add = model_instance._state.adding
            existing_path = None if add else self._get_existing_path(model_instance)

            # There are six scenarios to deal with:
            adding_blank = add and value is None
            adding_valid = add and value is not None
            updating_valid_to_valid = not add and self._get_valid_to_valid(model_instance)
            updating_valid_to_blank = not add and self._get_valid_to_blank(model_instance)
            updating_blank_to_valid = not add and self._get_blank_to_valid(model_instance)
            unchanged = not add and self._get_unchanged(model_instance)

            # Branch based on scenarios
            if unchanged:
                new_value = {"path": value["path"]} if value is not None else None

            elif adding_blank or updating_valid_to_blank:
                new_value = None

                # Trigger the on_change callback at the end of the commit when we know the
                # database transaction will work
                def on_commit_blank():
                    if self.on_change is not None:
                        self.on_change(new_value, instance=model_instance)

                # clean may happen outside transations so we must
                # actually register the transaction during pre-save,
                # store the callback here temporarily
                self._on_commit_blank = on_commit_blank

            elif adding_valid or updating_blank_to_valid or updating_valid_to_valid:
                new_value = {}

                allow_overwrite = self._get_allow_overwrite(add)

                attributes = self._update_attributes(
                    getattr(value, "attributes", {}),
                    instance=model_instance,
                    original_name=value["name"],
                    existing_path=existing_path,
                    temporary_path=value["_tmp_path"],
                    adding=add,
                    bucket=self.storage.bucket,
                )

                new_value["path"], allow_overwrite = self._get_destination_path(
                    instance=model_instance,
                    original_name=value["name"],
                    attributes=attributes,
                    allow_overwrite=self._get_allow_overwrite(add),
                    existing_path=existing_path,
                    temporary_path=value["_tmp_path"],
                    bucket=self.storage.bucket,
                )

                logger.info(
                    "Adding/updating cloud object via temporary ingress at %s to %s",
                    value["_tmp_path"],
                    new_value["path"],
                )

                # Trigger the copy only on successful commit of the transaction. We have to
                # capture the dual edge cases of the file not moving correctly, and the database
                # row not saving (eg due to validation errors in other model fields).
                # https://stackoverflow.com/questions/33180727/trigering-post-save-signal-only-after-transaction-has-completed
                def on_commit_valid():
                    copy_blob(
                        self.storage.bucket,
                        value["_tmp_path"],
                        self.storage.bucket,
                        new_value["path"],
                        move=True,
                        overwrite=allow_overwrite,
                        attributes=attributes,
                    )
                    if self.on_change is not None:
                        self.on_change(new_value, instance=model_instance)

                logger.info(
                    "Registered move of %s to %s to happen on transaction commit",
                    value["_tmp_path"],
                    new_value["path"],
                )
                self._on_commit_valid = on_commit_valid

            else:
                # Raise unknown edge cases rather than failing silently
                raise ValueError(
                    f"Unable to determine field state for {self._get_fieldname(model_instance)}. The most likely cause of this doing an operation (like migration) without setting GCP_STORAGE_OVERRIDE_BLOBFIELD_VALUE=True. Otherwise, please contact the django_gcp developers and describe what you're doing along with this exception stacktrace. Value was: {json.dumps(value)}"
                )

        # Cache DB values in the instance so you can reuse it without multiple DB queries
        # pylint: disable-next=protected-access
        model_instance._state.fields_cache[self.attname] = new_value

        self._cleaned = True

        return new_value

    def pre_save(self, model_instance, add):
        """Run on_commit hooks for transactions"""
        value = getattr(model_instance, self.attname)
        if not self._cleaned:
            value = self.clean(value, model_instance, skip_validation=True)

        if self._on_commit_blank is not None:
            transaction.on_commit(self._on_commit_blank)
            self._on_commit_blank = None

        if self._on_commit_valid is not None:
            transaction.on_commit(self._on_commit_valid)
            self._on_commit_valid = None

        # Reset the spaghetti used for meeting all django's awkward flows
        self._cleaned = False
        self._validated = False
        self._on_commit_blank = None
        self._on_commit_valid = None

        return value

    def validate(self, value, model_instance):
        """Validate field value contents
        Checks that the value is a dict and checks blankness.
        If the model is in an adding state, checks contents of the value and also presence
        of the ingressing file in the cloud store (this latter check causes an API request to the store)
        """

        # Override the superclass completely because it doesn't do anything useful
        # and has counterintuitive behaviour around what constitutes blank json
        # super().validate(value, model_instance)
        field_name = f"{model_instance.__class__.__name__}.{self.attname}"

        # Accept None or a dict
        value_or_dict = value or dict()
        if not isinstance(value_or_dict, dict):
            raise ValidationError(self.error_messages["invalid"], code="invalid", params={"value": value})

        if not self.blank and len(value_or_dict) == 0:
            raise ValidationError(self.error_messages["blank"], code="blank")

        has_tmp_path_or_name = "_tmp_path" in value_or_dict or "name" in value_or_dict
        has_tmp_path_and_name = "_tmp_path" in value_or_dict and "name" in value_or_dict
        has_path = "path" in value_or_dict
        is_blank = not has_path and not has_tmp_path_and_name
        # pylint: disable-next=protected-access
        adding = model_instance._state.adding
        updating = not adding
        blankable = self.blank
        editable = self.editable

        # Check that if a _tmp_path is given, a name is also given
        if has_tmp_path_or_name and not has_tmp_path_and_name:
            raise ValidationError(
                f"Both `_tmp_path` and `name` properties must be present in data for {field_name} if ingressing a new blob."
            )

        # Check the matrix of valid combinations with specific error messages for each
        if adding and blankable:
            # Valid values:
            #   None
            #   {"_tmp_path":"something-new", "name":"something"}
            if has_path:
                raise ValidationError("You cannot specify a path directly")
            if not is_blank and not has_tmp_path_and_name:
                raise ValidationError("Provide either a blank value or valid ingress location")

        elif adding and not blankable:
            # Valid values:
            #   {"_tmp_path":"something-new", "name":"something"}
            if has_path:
                raise ValidationError("You cannot specify a path directly")
            if not has_tmp_path_and_name:
                raise ValidationError("Provide a valid ingress location")

        elif updating and blankable and editable:
            # Valid values:
            #   None
            #   {"path":"something-unchanged-from-last"}
            #   {"_tmp_path":"something-new", "name":"something"}
            if not is_blank and not has_path and not has_tmp_path_and_name:
                raise ValidationError("Provide either a blank value, valid ingress location or the existing blob path")

        elif updating and blankable and not editable:
            # Valid values:
            #   None
            #   {"path":"something-unchanged-from-last"}
            if has_tmp_path_and_name:
                raise ValidationError("Cannot change an uneditable field")
            if not is_blank and not has_path:
                raise ValidationError("Field is not editable")

        elif updating and not blankable and editable:
            # Valid values:
            #   {"path":"something-unchanged-from-last"}
            #   {"_tmp_path":"something-new", "name":"something"}
            if not has_path and not has_tmp_path_and_name:
                raise ValidationError("Provide either a valid ingress location or the existing blob path")

        elif updating and not blankable and not editable:
            # Valid values:
            #   {"path":"something-unchanged-from-last"}
            if has_tmp_path_and_name:
                raise ValidationError("Cannot change an uneditable field")
            if not has_path:
                raise ValidationError("Provide the existing blob path")

        # Check for temporary blob completion
        if has_tmp_path_and_name:
            tmp_path = value["_tmp_path"]
            if not blob_exists(self.storage.bucket, tmp_path):
                raise ValidationError(
                    f"Upload incomplete or failed (no blob at '{tmp_path}' in bucket '{self.storage.bucket.name}')"
                )

        # Check that path roundtrips correctly without an attempt to change it
        if has_path:
            if not self.override_blobfield_value and self._get_path_altered(model_instance):
                raise ValidationError(
                    f"You cannot directly alter the `path` property in {field_name} data. Return the original `path` property to leave the field unchanged, or new `_tmp_path` and `name` properties to update the field."
                )

        self._validated = True

    @property
    def versioning_enabled(self):
        """True if object versioning is enabled on the bucket configured for this field"""
        if self._versioning_enabled is None:
            self._versioning_enabled = bool(self.storage.bucket.versioning_enabled)
        return self._versioning_enabled

    def _check_ingress_to(self):
        if isinstance(self.ingress_to, str) and self.ingress_to.startswith("/"):
            return [
                checks.Error(
                    "{self.__class__.__name__}'s 'ingress_to' argument must be a relative path, not an absolute path.",
                    obj=self,
                    id="fields.E202",
                    hint="Remove the leading slash.",
                )
            ]
        return []

    def _check_overwrite_mode(self):
        if self.overwrite_mode not in OVERWRITE_MODES:
            return [
                checks.Error(
                    "{self.__class__.__name__}'s 'overwrite_mode' is invalid",
                    obj=self,
                    id="fields.E202",
                    hint=f"Try one of: {OVERWRITE_MODES}",
                )
            ]
        return []

    def _check_explicit(self):
        if self._primary_key_set_explicitly or self._choices_set_explicitly:
            return [
                checks.Error(
                    f"'primary_key', and 'choices' are not valid arguments for a {self.__class__.__name__}",
                    obj=self,
                    id="fields.E201",
                )
            ]
        return []

    def _check_get_destination_path(self):
        if not callable(self.get_destination_path):
            return [
                checks.Error(
                    f"'get_destination_path' argument in {self.__class__.__name__} must be a callable function to return the destination object name.",
                    obj=self,
                    id="fields.E201",
                )
            ]
        return []

    def _check_update_attributes(self):
        if self.update_attributes is not None and not callable(self.update_attributes):
            return [
                checks.Error(
                    f"'update_attributes' argument in {self.__class__.__name__} must be None, or a callable function that updates attributes to be set on ingressed blobs.",
                    obj=self,
                    id="fields.E201",
                )
            ]
        return []

    def _check_on_change(self):
        if self.on_change is not None and not callable(self.on_change):
            return [
                checks.Error(
                    f"'on_change' argument in {self.__class__.__name__} must be or a callable function, or None",
                    obj=self,
                    id="fields.E201",
                )
            ]
        return []

    @staticmethod
    def _clean_blank_value(value):
        """Convert blanks values formulated as empty dicts {} to None"""
        if value is not None and len(value) == 0:
            value = None
        return value

    def _get_allow_overwrite(self, add):
        """Return a boolean determining if overwrite should be allowed for this operation"""
        mode_map = {
            "never": False,
            "update": not add,
            "update-versioned": self.versioning_enabled and not add,
            "add": add,
            "add-versioned": self.versioning_enabled and add,
            "add-update": True,
            "add-update-versioned": self.versioning_enabled,
        }
        return mode_map[self.overwrite_mode]

    def _get_blank_to_valid(self, instance):
        """Return true if overwriting a blank path with a valid one"""
        existing_path = self._get_existing_path(instance)
        temporary_path = self._get_instance_tmp_path(instance)
        return existing_path is None and temporary_path is not None

    def _get_existing_path(self, instance):
        """Gets the existing (saved in db) path; will raise an error if the instance is unsaved.
        Caches value on the instance to avoid duplicate database calls within a transaction.
        """
        # pylint: disable=protected-access
        if self.attname not in instance._state.fields_cache.keys():
            pk_name = instance._meta.pk.name
            # pylint: disable-next=protected-access
            existing = instance.__class__._default_manager.get(**{pk_name: getattr(instance, pk_name)})
            # Cache value in a dict because None is a valid value
            instance._state.fields_cache[self.attname] = getattr(existing, self.attname)
        value_or_dict = instance._state.fields_cache.get(self.attname, None) or {}
        return value_or_dict.get("path", None)

    def _get_fieldname(self, instance):
        """Return the ModelName.FieldName string for use in error messages and warnings"""
        return f"{instance.__class__.__name__}.{self.attname}"

    def _get_instance_path(self, instance):
        """Gets the instance path (or None) for a given instance"""
        value_or_dict = getattr(instance, self.attname) or {}
        return value_or_dict.get("path", None)

    def _get_instance_tmp_path(self, instance):
        """Gets the instance path (or None) for a given instance"""
        value_or_dict = getattr(instance, self.attname) or {}
        return value_or_dict.get("_tmp_path", None)

    def _get_path_altered(self, instance):
        """Return true if the path has changed between a given value and what's stored in the database.

        If there is no path in the database (e.g. a blank entry is stored) the result is False.

        Used only for validation purposes: the client side should never be able to specify the path.
        """
        existing_path = self._get_existing_path(instance)
        instance_path = self._get_instance_path(instance)

        return existing_path is not None and instance_path is not None and instance_path != existing_path

    def _get_signed_ingress_url(self):
        """Return a signed URL for uploading a blob to the temporary_path"""
        return get_signed_upload_url(
            self.storage.bucket,
            self._get_temporary_path(),
            content_type="application/octet-stream",
            max_size_bytes=self.max_size_bytes,
        )

    def _get_temporary_path(self):
        """Return a temporary path to which a blob can be uploaded before renaming"""
        if self._temporary_path is None:
            self._temporary_path = os.path.join(self.ingress_to, str(uuid4()))
        return self._temporary_path

    def _get_unchanged(self, instance):
        """Return true if the path value is simply returned unchanged with no _tmp_path"""
        existing_path = self._get_existing_path(instance)
        instance_path = self._get_instance_path(instance)
        temporary_path = self._get_instance_tmp_path(instance)
        return existing_path == instance_path and temporary_path is None

    def _get_valid_to_blank(self, instance):
        """Return true if overwriting a valid path with a blank one"""
        existing_path = self._get_existing_path(instance)
        temporary_path = self._get_instance_tmp_path(instance)
        return existing_path is not None and temporary_path is None

    def _get_valid_to_valid(self, instance):
        """Return true if updating object contents at an existing path"""
        existing_path = self._get_existing_path(instance)
        temporary_path = self._get_instance_tmp_path(instance)
        return existing_path is not None and temporary_path is not None

    def _get_destination_path(self, *args, **kwargs):
        """Call the get_destination_path callback unless an override is defined in
        settings. This funcitonality is intended for test purposes only, because
        patching the callback in a test framework is a struggle
        """
        get_destination_path = getattr(
            settings,
            "GCP_STORAGE_OVERRIDE_GET_DESTINATION_PATH_CALLBACK",
            self.get_destination_path,
        )
        return get_destination_path(*args, **kwargs)

    def _update_attributes(self, attributes, **kwargs):
        """Call the update_attributes callback unless an override is defined in
        settings. This funcitonality is intended for test purposes only, because
        patching the callback in a test framework is a struggle
        """
        update_attributes = getattr(
            settings,
            "GCP_STORAGE_OVERRIDE_UPDATE_ATTRIBUTES_CALLBACK",
            self.update_attributes,
        )
        if update_attributes is not None:
            return update_attributes(attributes, **kwargs)
        return attributes



import base64
from datetime import timezone
import json
import logging

from dateutil.parser import isoparse
from django.conf import settings
from django.urls import reverse
from django.utils.http import urlencode

from django_gcp.exceptions import InvalidPubSubMessageError

logger = logging.getLogger(__name__)


def _make_naive_utc(value):
    """Converts a timezone-aware datetime.datetime to UTC then makes it naive.
    Used for strictly formatting the UTC-in-nanoseconds publish time of pub/sub messages
    """
    if value.tzinfo is not None:
        value = value.astimezone(timezone.utc).replace(tzinfo=None)
    return value


def get_event_url(event_kind, event_reference, event_parameters=None, url_namespace="gcp-events", base_url=None):
    """Returns a fully constructed url for the events endpoint, suitable for receipt and processing of events
    :param str event_kind: The kind of the event (must be url-safe)
    :param str event_reference: A reference allowing either identification of unique events or a group of related events (must be url-safe)
    :param Union[dict, None] event_parameters: Dict of additional parameters to encode into the URL querystring, for example use {"token": "abc"} to add a token parameter that gets received by the endpoint.
    :param str url_namespace: Default 'gcp-events'. URL namespace of the django-gcp events (see https://docs.djangoproject.com/en/4.0/topics/http/urls/#url-namespaces)
    :param Union[str, None] base_url: The base url (eg https://somewhere.com) for the URL. By default, this uses django's BASE_URL setting. To generate an empty value (a relative URL) use an empty string ''.
    :return str: The fully constructed webhook endpoint
    """
    url = reverse(url_namespace, args=[event_kind, event_reference])
    if event_parameters is not None:
        url = url + "?" + urlencode(event_parameters)

    if base_url is None:
        try:
            base_url = settings.BASE_URL
        except AttributeError as e:
            raise AttributeError(
                "Either specify BASE_URL in your settings module, or explicitly pass a base_url parameter to get_push_endpoint()"
            ) from e

    url = base_url + url

    logger.debug("Generated webhook endpoint url %s", url)

    return url


def make_pubsub_message(
    data, subscription, attributes=None, message_id=None, ordering_key=None, publish_time=None, as_dict=False
):
    """Make a bytes object containing a message replicating the GCP Pub/Sub v1 format

    For more details see: https://cloud.google.com/pubsub/docs/reference/rest/v1/PubsubMessage

    Duplicated snake case keys found in "real" pubsub messages are not added, since they appear to be deprecated:
    https://stackoverflow.com/questions/73477187/why-do-gcp-pub-sub-messages-have-duplicated-message-id-and-publish-time-values

    :param Union[dict, list] data: JSON-serialisable data to form the body of the message
    :param str subscription: a subscription path which indicates the subscription that caused this message to be delivered, eg 'projects/my-project/subscriptions/my-subscription-name'
    :param Union[dict, None] attributes: Dict of attributes to attach to the message. Contents must be flat, containing only string keys with string values.
    :param Union[str, None] message_id: An optional id for the message.
    :param Union[str, None] ordering_key: A string used to order messages.
    :param Union[datetime, None] publish_time: If sending a message to PubSub, this will be set by the server on receipt so generally should be left as `None`. However, for the purposes of mocking messages for testing, supply a python datetime specifying the publish time of the message, which will be converted to a string timestamp with nanosecond accuracy.
    :param bool as_dict: If true, return the message as a dict instead of a utf-8-encoded json string
    :return bytes: A bytes object containing a fully composed PubSub message
    """
    message = dict()

    message["data"] = base64.b64encode(json.dumps(data).encode()).decode()

    if publish_time is not None:
        publish_time_utc_naive = _make_naive_utc(publish_time)
        iso_us = publish_time_utc_naive.isoformat()
        iso_ns = f"{iso_us}000Z"
        message["publishTime"] = iso_ns

    if attributes is not None:
        # Check all attributes are k-v pairs of strings
        for k, v in attributes.items():
            if k.__class__ != str:
                raise ValueError("All attribute keys must be strings")
            if v.__class__ != str:
                raise ValueError("All attribute values must be strings")
        message["attributes"] = attributes

    if message_id is not None:
        if message_id.__class__ != str:
            raise ValueError("The message_id, if given, must be a string")
        message["messageId"] = message_id

    if ordering_key is not None:
        if ordering_key.__class__ != str:
            raise ValueError("The ordering_key, if given, must be a string")
        message["orderingKey"] = ordering_key

    if subscription.__class__ != str:
        raise ValueError(
            "The subscription must be a string, like 'projects/my-project/subscriptions/my-subscription-name'"
        )

    if as_dict:
        return {"message": message, "subscription": subscription}

    return json.dumps({"message": message, "subscription": subscription}).encode("utf-8")


def decode_pubsub_message(body):
    """Decode data from a pubsub message body

    If the message data is json-decodable, then the decoded object's data field will contain the decoded object or array.
    Otherwise, the data field will contain a decoded string (pubsub messages comprising just a string are valid).

    :parameter Union[bytes, dict] body: The Pub/Sub message as a bytes object or as a decoded object
    :return: dict A flattened data structure containing message data and other information
    """
    if isinstance(body, bytes):
        body = json.loads(body)

    try:
        message = body["message"]
        subscription = body["subscription"]
        data = base64.b64decode(message["data"])

        # If data is json-decodable then do it. If it's just a string (which can be a valid message) accept that and decode it from bytes
        #   TODO determine if this try/catch is required. It was put there to handle raw strings coming in from pubsub, but unit tests show
        #   that the decoder will handle a single string
        try:
            data = json.loads(data)
        except json.decoder.JSONDecodeError:
            data = data.decode("utf-8")

        decoded = {
            "data": data,
            "attributes": message.get("attributes", None),
            "message_id": message.get("messageId", None),
            "ordering_key": message.get("orderingKey", None),
            "publish_time": message.get("publishTime", None),
            "subscription": subscription,
        }

    except KeyError as e:
        raise InvalidPubSubMessageError(
            f"Failed to decode Pub/Sub message (check the message conforms to Pub/Sub requirements: {body}"
        ) from e

    if decoded["publish_time"] is not None:
        decoded["publish_time"] = isoparse(decoded["publish_time"])

    return decoded



from django.template.loader import render_to_string
from django.utils.safestring import mark_safe

from .blob_utils import (
    get_console_url,
    get_path,
    get_signed_download_url,
)
from .fields import BlobField


class BlobFieldModelAdminMixin:
    def get_readonly_blob_widget(self, obj, field_name):
        context = {
            "existing_path": get_path(obj, field_name),
            "download_url": get_signed_download_url(obj, field_name),
            "console_url": get_console_url(obj, field_name),
        }
        return mark_safe(render_to_string("unfold/contrib/django_gcp/cloud_object_readonly_widget.html", context))

    def _replace_blob_field_names(self, fields, fields_to_replace):
        blob_fields = set(f.name for f in self.model._meta.get_fields() if isinstance(f, BlobField))

        # suffix field names with __readonly if they're blob fields
        for i, item in enumerate(fields):
            if item in fields_to_replace:
                fields[i] = f"{item}__readonly"

                # Check that the appropriate setup is done
                if item not in blob_fields:
                    raise ValueError(
                        f"The field {item} is not a BlobField, but is in readonly_blob_fields. "
                        f"Please remove it from readonly_blob_fields."
                    )
        return fields

    def get_fields(self, request, obj=None):
        fields = super().get_fields(request, obj)

        # Regardless of whether adding or editing, any read_only_blob_fields
        # need to be switched for custom defined fields
        if hasattr(self, "read_only_blob_fields"):
            fields = self._replace_blob_field_names(fields, self.read_only_blob_fields)

        if obj is not None and hasattr(self, "create_only_blob_fields"):
            # When adding a new object, create_only_blob_fields should be left alone
            # because they need to be set on object creation. When editing an existing
            # object, however, we ALSO pull the switcharoo create_only_blob_fields
            fields = self._replace_blob_field_names(fields, self.create_only_blob_fields)

        return fields

    def get_readonly_fields(self, request, obj=None):
        readonly_fields = super().get_readonly_fields(request, obj)

        if hasattr(self, "read_only_blob_fields"):
            readonly_fields = [*readonly_fields, *self.read_only_blob_fields]
            readonly_fields = self._replace_blob_field_names(readonly_fields, self.read_only_blob_fields)

        if obj is not None and hasattr(self, "create_only_blob_fields"):
            readonly_fields = [*readonly_fields, *self.create_only_blob_fields]
            readonly_fields = self._replace_blob_field_names(readonly_fields, self.create_only_blob_fields)

        return readonly_fields



from contextlib import contextmanager
import datetime
import logging

from django.test.utils import override_settings
from django.utils import timezone
from google.cloud.exceptions import NotFound, PreconditionFailed
from google.cloud.storage.blob import Blob

from django_gcp.exceptions import AttemptedOverwriteError, MissingBlobError

logger = logging.getLogger(__name__)

UNLIMITED_MAX_SIZE = 0


def blob_exists(bucket, blob_name):
    """Quick check that a blob with a given name exists in a bucket"""
    blob = bucket.blob(blob_name)
    return blob.exists()


def upload_blob(
    instance,
    field_name,
    local_path,
    destination_path=None,
    attributes=None,
    allow_overwrite=False,
    existing_path=None,
):
    """Upload a file to the cloud store, using the instance and field name to determine the store details

    You might use this utility to upload fixtue files for integration tests, as part of
    a migration, or as part of a function creating local files. The field's own logic for generating paths
    is used by default, although this can be overridden.

    Returns the field value so you can use this to construct instances directly:

    # Directly set the instance field without processing the blob ingress
    with override_settings(GCP_STORAGE_OVERRIDE_BLOBFIELD_VALUE=True):
        instance.field_name = upload_blob(...)

    :param django.db.Model instance: An instance of a django Model which has a BlobField
    :param str field_name: The name of the BlobField attribute on the instance
    :param str local_path: The path to the file to upload
    :param str destination_path: The path to upload the file to. If None, the remote path will be generated
    from the BlobField. If setting this value, take care to override the value of the field on the instance
    so it's path matches; this is not updated for you.
    :param dict attributes: A dictionary of attributes to set on the blob eg content type
    :param bool allow_overwrite: If true, allows existing blobs at the path to be overwritten. If destination_path is not given, this is provided to the get_destination_path callback (and may be overridden by that callback per its specification)
    :param str existing_path: If destination_path is None, this is provided to the get_destination_path callback to simulate behaviour where there is an existing path
    """
    # Get the field (which
    field = instance._meta.get_field(field_name)
    if destination_path is None:
        destination_path, allow_overwrite = field.get_destination_path(
            instance,
            original_name=local_path,
            attributes=attributes,
            existing_path=existing_path,
            temporary_path=None,
            allow_overwrite=allow_overwrite,
            bucket=field.storage.bucket,
        )

    # If not allowing overwrite, set generation matching constraints to prevent it
    if_generation_match = None if allow_overwrite else 0

    # Attributes must be a dict by default
    attributes = attributes or {}

    # Upload the file
    Blob(destination_path, bucket=field.storage.bucket).upload_from_filename(
        local_path, if_generation_match=if_generation_match, **attributes
    )

    # Return the field value
    return {"path": destination_path}


def copy_blob(
    source_bucket,
    source_blob_name,
    destination_bucket,
    destination_blob_name,
    overwrite=False,
    move=False,
    attributes=None,
):
    """Copies or moves a blob from one bucket to another with a new name.

    :param source_bucket: The google.cloud.storage.bucket object you're copying from
    :param source_blob_name: The name of the GCS object to be copied eg "path/in/bucket/source-blob-name.txt"
    :param destination_bucket_name: The google.cloud.storage.bucket object you're copying to
    :param destination_blob_name: The name of the new GCS object eg "path/in/bucket/destination-blob-name.txt"
    :param overwrite: If True, allows an existing destination file to be overwritten
    :param move: If True, removes the source blob after copy (ie moves it rather than duplicating it)
    :param attributes: A dict of values used to update attributes of the destination blob (eg {"content_type": "image/png"})
    """

    source_blob = source_bucket.blob(source_blob_name)

    extra_args = {}
    if not overwrite:
        # Optional: set a generation-match precondition to avoid potential race conditions
        # and data corruptions. The request is aborted if the object's
        # generation number does not match your precondition. For a destination
        # object that does not yet exist, set the if_generation_match precondition to 0.
        # If the destination object already exists in your bucket, set instead a
        # generation-match precondition using its generation number.
        # There is also an `if_source_generation_match` parameter, which is not used in this example.
        destination_generation_match_precondition = 0
        extra_args["if_generation_match"] = destination_generation_match_precondition

    try:
        destination_blob = source_bucket.copy_blob(source_blob, destination_bucket, destination_blob_name, **extra_args)

        verb = "copied"
        if move:
            verb = "moved"
            source_bucket.delete_blob(source_blob_name)

        logger.info(
            "Blob %s in bucket %s %s to blob %s in bucket %s",
            source_blob_name,
            source_bucket.name,
            verb,
            destination_blob_name,
            destination_bucket.name,
        )

        if attributes is not None:
            for key, value in attributes.items():
                setattr(destination_blob, key, value)
                logger.debug("Set attribute %s to %s on destination blob %s", key, value, destination_blob_name)

            destination_blob.patch()
            logger.info("Attributes of blob %s updated", destination_blob_name)

        return destination_blob

    except NotFound as e:
        raise MissingBlobError(
            f"Could not complete copy: source blob {source_blob_name} does not exist in bucket {source_bucket.name}"
        ) from e

    except PreconditionFailed as e:
        raise AttemptedOverwriteError(
            f"Could not complete copy: blob {destination_blob_name} in bucket {destination_bucket.name} already exists (source {source_blob_name} in bucket {source_bucket.name})"
        ) from e


def delete_blob(bucket, blob_name, generation=None, ignore_missing=False):
    """Deletes a blob, with the ability to handle missing blobs

    :param bucket: The bucket object from which the blob will be deleted
    :param blob_name: The name of the blob to be deleted eg "path/in/bucket/your-object-name.txt"
    :param generation: the specific generation of a versioned blob to delete. Unless generation is passed, deletion of versioned blobs will
    simply add a time_deleted parameter to those blobs.
    :param ignore_missing: If True, attempting to delete a missing blob will pass without raising an exception.
    This option should only be used when there's a real chance that objects may already have been cleaned up
    by another process by the time you call this (to prevent race conditions). Otherwise leave False since
    failure at this point can be indicative of upstream errors in handling that blob.
    """
    try:
        bucket.delete_blob(blob_name, generation=generation)
        logger.info("Deleted blob %s from bucket %s", blob_name, bucket.name)

    except NotFound as e:
        if not ignore_missing:
            raise MissingBlobError("Could not delete blob %s from bucket %s - blob not found") from e

        logger.info("Attempted to delete blob %s from bucket %s - blob missing", blob_name, bucket.name)


def get_generations(bucket, blob_name):
    """Get blobs corresponding to all generations of an object
    TODO Work up a more useful output than simply a list of objects
    :param bucket: The bucket object from which the blob will be deleted
    :param blob_name: The name of the blob to be deleted eg "path/in/bucket/your-object-name.txt"
    """

    return list(bucket.client.list_blobs(bucket.name, versions=True, prefix=blob_name))


def get_signed_upload_url(bucket, blob_name, timedelta=None, max_size_bytes=UNLIMITED_MAX_SIZE, **kwargs):
    """Get a signed URL for uploading a blob to GCS

    :param google.cloud.storage.Bucket bucket: The bucket to which the blob will be uploaded
    :param str blob_name: Name of the blob within the bucket
    :param Union[datetime.timedelta, None] timedelta: A datetime.timedelta object representing the
    period which the upload URL should be valid for (default is 60 minutes)
    :param Union[int, None] max_size_bytes: Maximum size allowable, added to any headers that are supplied in kwargs.
    """
    if timedelta is None:
        timedelta = datetime.timedelta(minutes=60)

    blob = bucket.blob(blob_name)

    if max_size_bytes != 0:
        content_length_range = f"0,{max_size_bytes}"
        headers = kwargs.pop("headers", {})
        headers["X-Goog-Content-Length-Range"] = content_length_range
        kwargs["headers"] = headers

    return blob.generate_signed_url(expiration=timezone.now() + timedelta, method="PUT", **kwargs)


@contextmanager
def uploaded_blob(instance, field_name, local_path, destination_path=None, allow_overwrite=False, delete_on_exit=False):
    """A context manager enabling the preparation of an instance with a blob uploaded from a local path.
    Optionally override the destination path (the ultimate path of the blob within the destination bucket) and allow_overwrite parameter.

    Usage:

    ```py
    my_instance = MyModel()
    with uploaded_blob(my_instance, 'my_field', '/path/to/local/file') as field_value
        my_instance.my_field = field_value
        my_instance.save()
    ```

    """

    with override_settings(GCP_STORAGE_OVERRIDE_BLOBFIELD_VALUE=True):
        field_value = upload_blob(
            instance,
            field_name=field_name,
            local_path=local_path,
            destination_path=destination_path,
            allow_overwrite=allow_overwrite,
        )
        try:
            yield field_value
        finally:
            if delete_on_exit:
                raise NotImplementedError("Deletion on exit is not yet implemented")



from datetime import datetime, timedelta

from django.core.management.base import BaseCommand

from django_gcp.storage import GoogleCloudStorage


# pylint: disable=missing-class-docstring
class Command(BaseCommand):
    help = "Cleanup temporary files in Google Cloud Storage. When ingressing files to temporary blobs, any failure to save the corresponding model will result in an orphaned upload."

    def add_arguments(self, parser):
        parser.add_argument("store_key", type=str, help="Google Cloud Storage key")
        parser.add_argument("--delete", action="store_true", help="Delete the temporary files")

    def handle(self, *args, **options):
        store_key = options["store_key"]
        delete_files = options["delete"]

        # Instantiate GoogleCloudStorage
        store = GoogleCloudStorage(store_key=store_key)
        bucket = store.bucket

        # Filter blobs with name starting with '_tmp' and age over 24 hours
        blob_list = bucket.list_blobs(prefix="_tmp")
        tmp_files = []
        for blob in blob_list:
            age = datetime.now() - blob.time_created.replace(tzinfo=None)
            if age > timedelta(hours=24):
                tmp_files.append(blob)

        # Print the list of temporary files
        if tmp_files:
            print(f"Temporary files to delete: {[blob.name for blob in tmp_files]}")
        else:
            print("No temporary files to delete")

        # Delete the temporary files if the --delete flag was passed
        if delete_files and tmp_files:
            bucket.delete_blobs(tmp_files)
            print(f"Deleted {len(tmp_files)} temporary files")



# TODO: Refactor tests to import members directly then remove this module export
from . import gcloud
from .blob_utils import BlobFieldMixin, get_blob, get_blob_name, get_path, get_signed_url
from .fields import BlobField
from .gcloud import GoogleCloudFile, GoogleCloudMediaStorage, GoogleCloudStaticStorage, GoogleCloudStorage
from .operations import upload_blob, uploaded_blob

__all__ = [
    "BlobField",
    "BlobFieldMixin",
    "GoogleCloudStorage",
    "GoogleCloudFile",
    "gcloud",
    "GoogleCloudMediaStorage",
    "GoogleCloudStaticStorage",
    "get_blob",
    "get_blob_name",
    "get_path",
    "get_signed_url",
    "upload_blob",
    "uploaded_blob",
]



import asyncio
import logging

from django.conf import settings

from django_gcp import exceptions

from . import tasks
from ._patch_cloud_scheduler import CloudScheduler
from ._pilot.pubsub import CloudSubscriber

logger = logging.getLogger(__name__)


# A note on the @property style of using django settings:
#
#   In many django libraries, it's commonplace to inspect and validate
#   the relevant django settings on start of the app, often adding them as
#   attributes to a controller or manager class like this TaskManager.
#
#   However, that's a much-misunderstood antipattern. The reason is that
#   in testing, the client app is instantiated on set up of the test, then
#   any variation in settings required for the test (i.e. applied with the
#   @override_settings decorator) is applied afterward... so while a
#   setting might change, any *copy* of it won't, so the test doesn't work.
#
#   There are two ways around this; either have a settings handler that
#   listens to the `setting_changed` signal, or directly access the relevant
#   setting each time. Here we do the latter (the storage module does the
#   former, as the storage settings are more complicated). By using
#   properties to get those settings, we have an easy access and a single
#   location where defaults are defined.


class TaskManager:
    """Manages and registers cloud tasks"""

    def __init__(self):
        self.on_demand_tasks = {}
        self.periodic_tasks = {}
        self.subscriber_tasks = {}

    @property
    def default_queue_name(self):
        """Return the GCP_TASKS_DEFAULT_QUEUE_NAME setting or a default"""
        return getattr(settings, "GCP_TASKS_DEFAULT_QUEUE_NAME")

    @property
    def delimiter(self):
        """Return the GCP_TASKS_DELIMITER setting or a default"""
        return getattr(settings, "GCP_TASKS_DELIMITER", "--")

    @property
    def disable_execute(self):
        """Return the GCP_TASKS_DISABLE_EXECUTE setting or default False"""
        return getattr(settings, "GCP_TASKS_DISABLE_EXECUTE", False)

    @property
    def domain(self):
        """Return the GCP_TASKS_DOMAIN setting or a default"""
        _domain = getattr(settings, "GCP_TASKS_DOMAIN")
        if not _domain.startswith("https://"):
            message = f"The GCP_TASKS_DOMAIN setting is invalid for use with a GCP PubSub push subscription. Endpoints need to be valid and secure (setting is: {_domain})."
            raise exceptions.InvalidEndpointError(message)
        return _domain

    @property
    def eager_execute(self):
        """Return the GCP_TASKS_EAGER_EXECUTE setting or a default"""
        return bool(getattr(settings, "GCP_TASKS_EAGER_EXECUTE", False))

    @property
    def region(self):
        """Return the GCP_TASKS_REGION setting or a default"""
        return getattr(settings, "GCP_TASKS_REGION", "europe-west1")

    @property
    def resource_affix(self):
        """Return the GCP_TASKS_RESOURCE_AFFIX setting or a default"""
        return getattr(settings, "GCP_TASKS_RESOURCE_AFFIX", None)

    def register_task(self, task_class):
        """Record the presence of a task class in this manager
        Allows iteration through different task classes for maintenance operations.
        """

        # TODO REFACTOR REQUEST There's ambiguity in what "register" means. Here, register means
        # "add to a dict where we record all the tasks" which is probably the correct one.
        # But, the Task class has a `register` method which has varying functionality depending on
        # how it is overloaded, and can mean "create something on google cloud". So it's necessary to
        # extract those meanings, and probably rename the method in Task(), e.g. to `create_resources`
        # or similar

        name = task_class.name()

        if not getattr(task_class, "abstract", False):
            if issubclass(task_class, tasks.SubscriberTask):
                logger.debug("Registering Subscriber Task %s", name)
                self.subscriber_tasks[name] = task_class

            elif issubclass(task_class, tasks.PeriodicTask):
                logger.debug("Registering Periodic Task %s", name)
                self.periodic_tasks[name] = task_class

            elif issubclass(task_class, tasks.OnDemandTask):
                logger.debug("Registering On Demand Task %s", name)
                self.on_demand_tasks[name] = task_class
        else:
            logger.debug("Skipping registration of Task %s (because the task class is abstract)", name)

    def create_scheduler_jobs(self, cleanup=False):
        """Create a scheduler jobs in GCP for each PeriodicTask

        Any existing scheduler jobs with the resource prefix will be removed

        """
        updated = []
        removed = []

        # TODO REFACTOR REQUEST In many places, 'scheduler jobs' (the actual name of the GCP resource)
        # are referred to as 'schedules' or similar. Collapse all terminology onto 'scheduler jobs'
        # so it's more intuitive to retermine what the resource is.

        # TODO REFACTOR REQUEST It doesn't make sense that the client usage is split between
        # here (for removing outdated scheduler jobs) and the task.schedule() method for creating
        # scheduler jobs. Either add a `clean_resources` class method to the task to do the cleanup
        # or move the creation code here.

        for _, task_klass in self.periodic_tasks.items():
            task = task_klass()
            task.schedule()
            updated.append(task.schedule_name)

        if cleanup:
            if self.resource_affix:
                logger.debug("Cleaning up unused Scheduler Jobs")
                client = CloudScheduler(location=self.region)
                for job in client.list(prefix=self.resource_affix):
                    schedule_name = job.name.split("/jobs/")[-1]
                    if schedule_name not in updated:
                        asyncio.run(client.delete(name=schedule_name))
                        removed.append(schedule_name)
            else:
                logger.warning(
                    "Cleanup of unused Scheduler Jobs was requested but skipped, because no resource affix was given - see the GCP_TASKS_RESOURCE_AFFIX docs for more details."
                )
        else:
            logger.debug("Skipping cleanup of unused Scheduler Jobs")

        return updated, removed

    def create_pubsub_subscriptions(self, cleanup=False):
        """Register all SubscriberTasks by creating push subscriptions for them on PubSub

        Any existing PubSub subscriptions with the resource prefix will be removed from PubSub.

        """
        updated = []
        removed = []

        # TODO REFACTOR REQUEST It doesn't make sense that the client usage is split between
        # here (for removing outdated subscriptions) and the task.register() method for creating
        # subscriptions. Either add a `clean_resources` class method to the task to do the cleanup
        # or move the creation code here.

        for task_name, task_klass in self.subscriber_tasks.items():
            task_klass().register()
            updated.append(task_name)

        async def _get_subscriptions():
            names = []
            async for subscription in client.list_subscriptions(prefix=self.resource_affix):
                susbcription_name = subscription.name.rsplit("subscriptions/", 1)[-1]
                task_name = subscription.push_config.push_endpoint.rsplit("/", 1)[-1]
                names.append((susbcription_name, task_name))
            return names

        if cleanup:
            if self.resource_affix:
                logger.debug("Cleaning up unused PubSub Subscriptions")
                client = CloudSubscriber()
                for subscription_id, subscribed_task in asyncio.run(_get_subscriptions()):
                    if subscribed_task not in updated:
                        asyncio.run(client.delete_subscription(subscription_id=subscription_id))
                        removed.append(subscribed_task)
            else:
                logger.warning(
                    "Cleanup of unused PubSub Subscriptions was requested but skipped, because no resource affix was given - see the GCP_TASKS_RESOURCE_AFFIX docs for more details."
                )

        else:
            logger.debug("Skipping cleanup of unused PubSub Subscriptions")

        return updated, removed



from .base import _CACHED_LOCATIONS, AppEngineBasedService, DiscoveryMixin, GoogleCloudPilotAPI


class AppEngine(AppEngineBasedService, DiscoveryMixin, GoogleCloudPilotAPI):
    _scopes = [
        "https://www.googleapis.com/auth/appengine.admin",
    ]

    def __init__(self, **kwargs):
        super().__init__(
            serviceName="appengine",
            version="v1",
            cache_discovery=False,
            **kwargs,
        )

    def _set_location(self, location: str = None) -> str:
        return location or self._get_default_location()

    def get_app(self, app_id: str = None):
        app_id = app_id or self.project_id
        return self._execute(
            method=self.client.apps().get,
            appsId=app_id,
        )

    def _get_default_location(self, default_zone: str = "1") -> str:
        location = _CACHED_LOCATIONS.get(self.project_id, None)

        if not location:
            app = self.get_app()
            location = app["locationId"]
            try:
                int(location[-1])
            except ValueError:
                location = app["locationId"] + default_zone
            _CACHED_LOCATIONS[self.project_id] = location
        return location


__all__ = ("AppEngine",)



# Disabled while refactoring:
# pylint: disable=missing-docstring
# pylint: disable=missing-function-docstring
# pylint: disable=no-member
from abc import abstractmethod
from datetime import datetime, timedelta
import hashlib
import logging

from django.apps import apps
from django.template.defaultfilters import slugify
from django.urls import reverse
from django.utils.timezone import now
from google.api_core.exceptions import AlreadyExists
from google.cloud import pubsub_v1

from django_gcp.events.utils import decode_pubsub_message
from django_gcp.exceptions import DuplicateTaskError, IncompatibleSettingsError, IncorrectTaskUsageError

from ._patch_cloud_scheduler import CloudScheduler
from ._patch_cloud_tasks import CloudTasks
from ._pilot.pubsub import CloudPublisher, CloudSubscriber
from .helpers import run_coroutine
from .serializers import deserialize, serialize

logger = logging.getLogger(__name__)


def apply_resource_affix(value, suffix=False):
    manager = apps.get_app_config("django_gcp").task_manager
    if manager.resource_affix is not None:
        if suffix:
            return f"{value}{manager.delimiter}{manager.resource_affix}"
        else:
            return f"{manager.resource_affix}{manager.delimiter}{value}"

    return value


def apply_prefix(value):
    return apply_resource_affix(value, suffix=False)


def apply_suffix(value):
    return apply_resource_affix(value, suffix=True)


def short_sha(value, digits=8):
    """Calculate a short sha from an input string, with a certain degree of uniqueness"""
    return hashlib.sha256(value.encode("utf-8")).hexdigest()[:digits]


class TaskMeta(type):
    """Metadata class representing a Task"""

    def __new__(cls, name, bases, attrs):
        """On creation of a new Task subclass, register it with the task manager for the django-gcp app"""
        klass = type.__new__(cls, name, bases, attrs)

        # TODO Simplify this statement, what on earth is its purpose?
        #      If the class has attr abstract=True but abstract not in attrs set abstract=False?? WTF??
        if getattr(klass, "abstract", False) and "abstract" not in attrs:
            setattr(klass, "abstract", False)

        if klass.__name__ not in ["Task", "OnDemandTask", "PeriodicTask", "SubscriberTask"]:
            apps.get_app_config("django_gcp").task_manager.register_task(task_class=klass)

        return klass

    def __call__(cls, *args, **kwargs):
        """Override a task's __call__ method to check that it's been appropriately subclassed
        This catches the most common usage error.
        """
        if cls.__name__ in ["Task", "OnDemandTask", "PeriodicTask", "SubscriberTask"]:
            raise IncorrectTaskUsageError(f"Do not instantiate a {cls.__name__}. Inherit and create your own.")

        return super().__call__(*args, **kwargs)


class Task(metaclass=TaskMeta):
    """Task base class"""

    _url_name = "gcp-tasks"
    deduplicate = False

    def enqueue(self, **kwargs):
        """Invoke a task (place it onto a queue for processing)

        Some settings affect the behaviour of this method, allowing bypass or immediate execution. See settings documentation for more.
        """
        return self._send(
            task_kwargs=kwargs,
        )

    def enqueue_later(self, when, **kwargs):
        """Invoke a task (place it onto a queue for processing after some time delay)

        Some settings affect the behaviour of this method, allowing bypass or immediate execution. See settings documentation for more.
        """
        if isinstance(when, int):
            delay_in_seconds = when
        elif isinstance(when, timedelta):
            delay_in_seconds = when.total_seconds()
        elif isinstance(when, datetime):
            delay_in_seconds = (when - now()).total_seconds()
        else:
            raise ValueError(f"Unsupported schedule {when} of type {when.__class__.__name__}")

        return self._send(
            task_kwargs=kwargs,
            api_kwargs=dict(delay_in_seconds=delay_in_seconds),
        )

    def execute(self, request_body):
        """Deserialises the received request and calls the run() method"""
        try:
            task_kwargs = self._body_to_kwargs(request_body=request_body)
        except Exception as e:
            logger.warning(e, exc_info=True)
            return f"Unable to parse request arguments. Error was: {e}", 400

        try:
            return self.run(**task_kwargs), 200
        except Exception as e:
            logger.error(e, exc_info=True)
            return "Error running task", 500

    @property
    def manager(self):
        """The task manager instance used to record registered apps"""
        return apps.get_app_config("django_gcp").task_manager

    @classmethod
    def name(cls):
        """Name of the task class"""
        return cls.__name__

    @property
    def queue_name(self):
        """Name of the queue that tasks will be placed on
        Override to use a different queue for different task subclasses
        """
        return self.manager.default_queue_name

    @property
    def slug(self):
        """Slugified name of this class"""
        return slugify(self.name())

    @abstractmethod
    def run(self, **kwargs):
        raise NotImplementedError()

    @classmethod
    def url(cls):
        domain = apps.get_app_config("django_gcp").task_manager.domain
        path = reverse(cls._url_name, args=(cls.name(),))
        return f"{domain}{path}"

    def _body_to_kwargs(self, request_body):
        data = deserialize(request_body)
        return data

    # TODO REFACTOR REQUEST The base `Task` class seems to deal with on-demand tasks
    # (using Cloud Tasks) then is inherited by other classes using different task
    # mechanisms (e.g. Cloud Scheduler and Pub/Sub).
    # The Task class is thus treated only partly as an ABC.
    # We should refactor Task to split it into two; a true ABC containing only
    # common functionality, and an OnDemandTask class which implements the detail of
    # interacting with Cloud Tasks, like the following _send method

    def _send(self, task_kwargs, api_kwargs=None):
        if self.manager.disable_execute and self.manager.eager_execute:
            raise IncompatibleSettingsError("disable_execute and eager_execute should be mutually exclusive")

        if self.manager.disable_execute:
            return None

        payload = serialize(task_kwargs)
        if self.manager.eager_execute:
            return self.run(**deserialize(payload))

        api_kwargs = api_kwargs or {}
        api_kwargs.update(
            dict(
                queue_name=self.queue_name,
                url=self.url(),
                payload=payload,
            )
        )

        if self.deduplicate:
            # NOTE:
            #   If deduplicating tasks, a task ID must be supplied. As noted here in the GCP docs,
            #   this creates significant additional latency into the task allocation. Efficient
            #   operation of the task queue demands a binomially distributed task ID, meaning
            #   that we don't want to be adding a uniform resource prefix; so we use a suffix instead.
            #   The short sha prefix is used to "uniquely" (within reason) identify payloads, meaning you
            #   can trigger the same task repeatedly with different payloads, but attempting to repeat a
            #   task with the same payload will fail for ~1hour after the original task was deleted or
            #   executed.
            unique_task_name = apply_suffix(f"{short_sha(payload)}{self.manager.delimiter}{self.slug}")
            api_kwargs.update(
                dict(
                    task_name=unique_task_name,
                    unique=False,
                )
            )

            try:
                return run_coroutine(handler=self.__client.push, **api_kwargs)
            except AlreadyExists as e:
                raise DuplicateTaskError(
                    "Duplicate task detected (task already exists with this name and payload sha). You can trigger the same task repeatedly with different payloads, but attempting to repeat a task with the same payload will fail for ~1hour after the original task was deleted or executed."
                ) from e
        else:
            return run_coroutine(handler=self.__client.push, **api_kwargs)

    @property
    def __client(self):
        return CloudTasks(location=self.manager.region)


class OnDemandTask(Task):
    """Inherit from here to enable on-demand tasks

    This class is empty right now, but is here to avoid a breaking change later when we refactor
    the Task class. You should inherit from this.

    """

    abstract = True

    # TODO refactor the functionality specific to on-demand tasks to here from the Task class, because
    #  currently the Task class functions both as a base and as a specific variant class.

    @abstractmethod
    def run(self, **kwargs):
        raise NotImplementedError()


class PeriodicTask(Task):
    run_every = None

    @abstractmethod
    def run(self, **kwargs):
        raise NotImplementedError()

    def schedule(self, **kwargs):
        payload = serialize(kwargs)

        if self.manager.eager_execute:
            return self.run(**deserialize(payload))

        return run_coroutine(
            handler=self.__client.put,
            name=self.schedule_name,
            url=self.url(),
            payload=payload,
            cron=self.run_every,
        )

    @property
    def schedule_name(self):
        return apply_prefix(self.slug)

    @property
    def __client(self):
        return CloudScheduler(location=self.manager.region)


class SubscriberTask(Task):
    abstract = True
    _use_oidc_auth = True
    _url_name = "gcp-subscriber-tasks"
    enable_message_ordering = False

    def publish(self, data, attributes=None):
        """Publish a message onto the PubSub topic that this subscriber listens to

        This is mostly a convenience method for running integration tests, it's not
        intended for general use (What would be the point of sending a message to
        yourself? It's possible to use Pub/Sub as a tasks queue this way but that
        doesn't seem like it'd ever be useful compared to using the purpose-built
        cloud tasks queue).

        """
        return run_coroutine(
            handler=self.__publisher_client.publish,
            message=serialize(data),
            topic_id=self.topic_id,
            attributes=attributes,
        )

    def register(self):
        return run_coroutine(
            handler=self.__client.create_or_update_subscription,
            topic_id=self.topic_id,
            subscription_id=self.subscription_id,
            enable_message_ordering=self.enable_message_ordering,
            push_to_url=self.url(),
            use_oidc_auth=self._use_oidc_auth,
        )

    @abstractmethod
    def run(self, data, attributes, message_id, ordering_key, publish_time, subscription, **kwargs):  # pylint: disable=arguments-differ
        raise NotImplementedError()

    @property
    def subscription_id(self):
        return apply_prefix(f"{self.topic_id}{self.manager.delimiter}{self.slug}")

    @property
    @abstractmethod
    def topic_id(self):
        raise NotImplementedError()

    def _body_to_kwargs(self, request_body):
        return decode_pubsub_message(request_body)

    @property
    def __client(self):
        return CloudSubscriber()

    @property
    def __publisher_client(self):
        publisher_options = pubsub_v1.types.PublisherOptions(
            enable_message_ordering=self.enable_message_ordering,
        )
        return CloudPublisher(publisher_options=publisher_options)



import logging
import mimetypes
from tempfile import SpooledTemporaryFile

from django.conf import settings
from django.core.exceptions import SuspiciousOperation
from django.core.files.base import File
from django.core.files.storage import Storage
from django.utils import timezone
from django.utils.deconstruct import deconstructible
from google.cloud.exceptions import NotFound
from google.cloud.storage import Blob, Client
from google.cloud.storage.blob import _quote

from .compress import CompressedFileMixin, CompressStorageMixin
from .settings import StorageSettings
from .utils import clean_name, get_available_overwrite_name, safe_join, to_bytes

CONTENT_ENCODING = "content_encoding"
CONTENT_TYPE = "content_type"


logger = logging.getLogger(__name__)


class GoogleCloudFile(CompressedFileMixin, File):
    """A django File object representing a GCP storage object"""

    def __init__(self, name, mode, storage):  # pylint: disable=super-init-not-called
        self.name = name
        self.mime_type = mimetypes.guess_type(name)[0]
        self._mode = mode
        self._storage = storage
        self.blob = storage.bucket.get_blob(name)
        if not self.blob and "w" in mode:
            self.blob = Blob(self.name, storage.bucket, chunk_size=storage.settings.blob_chunk_size)
        self._file = None
        self._is_dirty = False

    def size(self):
        return self.blob.size

    def _get_file(self):
        if self._file is None:
            self._file = SpooledTemporaryFile(
                max_size=self._storage.settings.max_memory_size,
                suffix=".GSStorageFile",
                dir=getattr(settings, "FILE_UPLOAD_TEMP_DIR"),
            )
            if "r" in self._mode:
                self._is_dirty = False
                self.blob.download_to_file(self._file)
                self._file.seek(0)
            if self._storage.settings.gzip and self.blob.content_encoding == "gzip":
                self._file = self._decompress_file(mode=self._mode, file=self._file)
        return self._file

    def _set_file(self, value):
        self._file = value

    file = property(_get_file, _set_file)

    def read(self, num_bytes=None):
        """Read from the file-like object"""
        if "r" not in self._mode:
            raise AttributeError("File was not opened in read mode.")

        if num_bytes is None:
            num_bytes = -1

        return super().read(num_bytes)

    def write(self, content):
        """Write to the file-like object"""
        if "w" not in self._mode:
            raise AttributeError("File was not opened in write mode.")
        self._is_dirty = True
        return super().write(to_bytes(content))

    def close(self):
        """Close the file-like object"""
        if self._file is not None:
            if self._is_dirty:
                blob_params = self._storage.get_object_parameters(self.name)
                self.blob.upload_from_file(
                    self.file,
                    rewind=True,
                    content_type=self.mime_type,
                    predefined_acl=blob_params.get("acl", self._storage.settings.default_acl),
                )
            self._file.close()
            self._file = None


# Note on disabling plyint's abstract-method:
#   "Method 'path' is abstract in class 'Storage' but is not overridden"
#   But, collectstatic checks presence of the path method on this class,
#   so we can't trivially override it. See
#      - https://github.com/octue/django-gcp/releases/tag/0.7.3
#      - https://github.com/octue/django-gcp/issues/21
#
@deconstructible
class GoogleCloudStorage(CompressStorageMixin, Storage):  # pylint: disable=abstract-method
    """Storage class allowing django to use GCS as an object store

    Instantiates as the `media` store by default.
    Pass `media`, `static` or any of the keys in.
    """

    def __init__(self, store_key="media", **overrides):
        super().__init__()

        self.settings = StorageSettings(store_key, **overrides)
        self._bucket = None
        self._client = None

    def get_accessed_time(self, *_, **__):
        """Get the last accessed time of the file

        This value is ALWAYS None because we cannot know the last accessed time on a cloud file.

        This method is here for API compatibility with django's Storage class.
        """
        return None

    @property
    def client(self):
        """The google-storage client for this store"""
        if self._client is None:
            self._client = Client(project=self.settings.project_id, credentials=self.settings.credentials)
        return self._client

    @property
    def bucket(self):
        """The google-storage bucket object for this store
        See bucket class information: https://cloud.google.com/python/docs/reference/storage/latest/google.cloud.storage.bucket.Bucket
        """
        if self._bucket is None:
            self._bucket = self.client.bucket(self.settings.bucket_name)
        return self._bucket

    @property
    def versioning_enabled(self):
        """True if versioning is enabled on the bucket
        https://cloud.google.com/python/docs/reference/storage/latest/google.cloud.storage.bucket.Bucket#google_cloud_storage_bucket_Bucket_versioning_enabled
        """
        return self.bucket.versioning_enabled

    @property
    def bucket_name(self):
        """The name of the bucket corresponding to this store

        A convenience property referring to self.settings.bucket_name but here for backward compatibility
        with django_storages and ease of general use
        """
        return self.settings.bucket_name

    def _normalize_name(self, name):
        """
        Normalizes the name so that paths like /path/to/ignored/../something.txt
        and ./file.txt work.  Note that clean_name adds ./ to some paths so
        they need to be fixed here. We check to make sure that the path pointed
        to is not outside the directory specified by the LOCATION setting.
        """
        try:
            return safe_join(self.settings.location, name)
        except ValueError as e:
            logger.warning(e, stack_info=True)
            # Disable pylint recommendation to avoid attaching further information to any SuspiciousOperation
            raise SuspiciousOperation(f"Attempted access to '{name}' denied.")  # pylint: disable=raise-missing-from

    def _open(self, name, mode="rb"):
        name = self._normalize_name(clean_name(name))
        file_object = GoogleCloudFile(name, mode, self)
        if not file_object.blob:
            raise FileNotFoundError(f"File does not exist: {name}")
        return file_object

    def _save(self, name, content):
        cleaned_name = clean_name(name)
        name = self._normalize_name(cleaned_name)
        content.name = cleaned_name
        file_object = GoogleCloudFile(name, "rw", self)

        upload_params = {}
        blob_params = self.get_object_parameters(content)
        upload_params["predefined_acl"] = blob_params.pop("acl", self.settings.default_acl)
        upload_params[CONTENT_TYPE] = blob_params.pop(CONTENT_TYPE, file_object.mime_type)

        if (
            self.settings.gzip
            and upload_params[CONTENT_TYPE] in self.settings.gzip_content_types
            and CONTENT_ENCODING not in blob_params
        ):
            content = self._compress_content(content)
            blob_params[CONTENT_ENCODING] = "gzip"

        for prop, val in blob_params.items():
            setattr(file_object.blob, prop, val)

        file_object.blob.upload_from_file(content, rewind=True, size=getattr(content, "size", None), **upload_params)
        return cleaned_name

    def get_object_parameters(self, content):
        """Add object-specific parameters to the uploaded blob.

        Override this to return a dictionary of blob-specific parameters for the upload.

        By default, this uses any object_parameters configured on the store, except:

            - A `custom_time` attribute on the `content` object will override a `custom_time` property in the store settings
            - Any `metadata` on the `content` object will be merged with any `metadata` in the store settings, with fields
              from the `content` object taking precedence over the store metadata where duplicate fields exist.

        """
        object_parameters = self.settings.object_parameters.copy()
        store_metadata = getattr(object_parameters, "metadata", dict())
        store_custom_time = getattr(object_parameters, "custom_time", None)
        object_metadata = getattr(content, "metadata", dict())
        object_parameters["custom_time"] = getattr(content, "custom_time", store_custom_time)
        object_parameters["metadata"] = {**store_metadata, **object_metadata}

        return object_parameters

    def delete(self, name):
        name = self._normalize_name(clean_name(name))
        try:
            self.bucket.delete_blob(name)
        except NotFound:
            pass

    def exists(self, name):
        if not name:  # root element aka the bucket
            try:
                self.client.get_bucket(self.bucket)
                return True
            except NotFound:
                return False

        name = self._normalize_name(clean_name(name))
        return bool(self.bucket.get_blob(name))

    def listdir(self, path):
        prefix = self._normalize_name(clean_name(path))
        # For bucket.list_blobs and logic below name needs to end in /
        # but for the root path "" we leave it as an empty string
        if prefix and not prefix.endswith("/"):
            prefix += "/"

        iterator = self.bucket.list_blobs(prefix=prefix, delimiter="/")
        blobs = list(iterator)
        prefixes = iterator.prefixes

        files = []
        dirs = []

        for blob in blobs:
            parts = blob.name.split("/")
            files.append(parts[-1])
        for folder_path in prefixes:
            parts = folder_path.split("/")
            dirs.append(parts[-2])

        return list(dirs), files

    def _get_blob(self, name):
        # Wrap google.cloud.storage's blob to raise if the file doesn't exist
        blob = self.bucket.get_blob(name)

        if blob is None:
            raise NotFound(f"File does not exist: {name}")

        return blob

    def size(self, name):
        name = self._normalize_name(clean_name(name))
        blob = self._get_blob(name)
        return blob.size

    def modified_time(self, name):
        name = self._normalize_name(clean_name(name))
        blob = self._get_blob(name)
        return timezone.make_naive(blob.updated)

    def get_modified_time(self, name):
        name = self._normalize_name(clean_name(name))
        blob = self._get_blob(name)
        updated = blob.updated
        return updated if getattr(settings, "USE_TZ") else timezone.make_naive(updated)

    def get_created_time(self, name):
        """
        Return the creation time (as a datetime) of the file specified by name.
        The datetime will be timezone-aware if USE_TZ=True.
        """
        name = self._normalize_name(clean_name(name))
        blob = self._get_blob(name)
        created = blob.time_created
        return created if getattr(settings, "USE_TZ") else timezone.make_naive(created)

    def url(self, name):
        """
        Return public url or a signed url for the Blob.
        This DOES NOT check for existance of Blob - that makes codes too slow
        for many use cases.
        """
        name = self._normalize_name(clean_name(name))
        blob = self.bucket.blob(name)
        blob_params = self.get_object_parameters(name)
        no_signed_url = (
            blob_params.get("acl", self.settings.default_acl) == "publicRead" or not self.settings.querystring_auth
        )

        if not self.settings.custom_endpoint and no_signed_url:
            return blob.public_url
        elif no_signed_url:
            return "{storage_base_url}/{quoted_name}".format(
                storage_base_url=self.settings.custom_endpoint,
                quoted_name=_quote(name, safe=b"/~"),
            )
        elif not self.settings.custom_endpoint:
            return blob.generate_signed_url(expiration=self.settings.expiration, version="v4")
        else:
            return blob.generate_signed_url(
                bucket_bound_hostname=self.settings.custom_endpoint,
                expiration=self.settings.expiration,
                version="v4",
            )

    def get_available_name(self, name, max_length=None):
        name = clean_name(name)
        if self.settings.file_overwrite:
            return get_available_overwrite_name(name, max_length)
        return super().get_available_name(name, max_length)


class GoogleCloudMediaStorage(GoogleCloudStorage):  # pylint: disable=abstract-method
    """Storage whose bucket name is taken from the GCP_STORAGE_MEDIA_NAME setting

    This actually behaves exactly as a default instantiation of the base
    ``GoogleCloudStorage`` class, but is there to make configuration more
    explicit for first-timers.

    """

    def __init__(self, **overrides):
        if overrides.pop("store_key", "media") != "media":
            raise ValueError("You cannot instantiate GoogleCloudMediaStorage with a store_key other than 'media'")
        super().__init__(store_key="media", **overrides)


class GoogleCloudStaticStorage(GoogleCloudStorage):  # pylint: disable=abstract-method
    """Storage defined with an appended bucket name (called "<bucket>-static")

    We define that static files are stored in a different bucket than the (private) media files, which:
        1. gives us less risk of accidentally setting private files as public
        2. allows us easier visual inspection in the console of what's private and what's public static
        3. allows us to set blanket public ACLs on the static bucket
        4. makes it easier to clean up private files with no entry in the DB

    """

    def __init__(self, **overrides):
        if overrides.pop("store_key", "static") != "static":
            raise ValueError("You cannot instantiate GoogleCloudStaticStorage with a store_key other than 'static'")
        super().__init__(store_key="static", **overrides)



import abc

from django.apps import apps
from django.core.management.base import BaseCommand as DjangoBaseCommand


class BaseCommand(DjangoBaseCommand, abc.ABC):
    """A BaseCommand with additional helpers for task-related management commands"""

    @property
    def task_manager(self):
        """Return the task manager via the app configuration"""
        return apps.get_app_config("django_gcp").task_manager

    def display_task_report(self, report, action, name):
        """Print a report to stdout on the task successes"""

        n = len(report)
        report_str = "\n".join([f"- {name}" for name in report])

        message = f"Successfully {action}d {n} {name} to domain {self.task_manager.domain}\n{report_str}"
        self.stdout.write(self.style.SUCCESS(message))  # pylint: disable=no-member



from django.urls import path

from django_gcp.events.views import GoogleCloudEventsView
from django_gcp.tasks.views import GoogleCloudSubscriberTaskView, GoogleCloudTaskView

urlpatterns = [
    path(r"events/<event_kind>/<event_reference>", GoogleCloudEventsView.as_view(), name="gcp-events"),
    path(r"subscriber-tasks/<task_name>", GoogleCloudSubscriberTaskView.as_view(), name="gcp-subscriber-tasks"),
    path(r"tasks/<task_name>", GoogleCloudTaskView.as_view(), name="gcp-tasks"),
]
